/*************************************************************/
/*                                                           */
/*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

#ifdef OSTYPE_irix
#include <alloca.h>
#endif
#ifdef OSTYPE_solaris
#include <alloca.h>
#endif
#ifdef OSTYPE_linux
#include <alloca.h>
#endif

/*************************************************************/
/* structures                                                */
/*************************************************************/

#define FALSE 0
#define TRUE  1

/*#define BLAS_FLOPS_CUTOFF  1000.0*/
#define BLAS_FLOPS_CUTOFF  -1.0
#define SOLVE_DENSE_CUTOFF 5

typedef struct {
  int     sn_size;
  int     n;
  int*    rowind;

  int     up_size;
  int*    sn_vertices;
  int*    up_vertices;

  taucs_datatype* f1;
  taucs_datatype* f2;
  taucs_datatype* u;

} supernodal_frontal_matrix;

#define SFM_F1 f1
#define SFM_F2 f2
#define SFM_U   u

typedef struct {
  int     flags;

  char    uplo;     /* 'u' for upper, 'l' for lower, ' ' don't know; prefer lower. */
  int     n;        /* size of matrix */
  int     n_sn;     /* number of supernodes */

  int* parent;      /* supernodal elimination tree */
  int* first_child; 
  int* next_child;

  int* sn_size;     /* size of supernodes (diagonal block) */
  int* sn_up_size;  /* size of subdiagonal update blocks   */
  int** sn_struct;  /* row structure of supernodes         */

  int* sn_blocks_ld;  /* lda of supernode blocks */
  taucs_datatype** sn_blocks; /* supernode blocks        */
    
  int* up_blocks_ld;  /* lda of update blocks    */
  taucs_datatype** up_blocks; /* update blocks           */
} supernodal_factor_matrix;

/*************************************************************/
/* for qsort                                                 */
/*************************************************************/

static int compare_ints(void* vx, void* vy)
{
  int* ix = (int*)vx;
  int* iy = (int*)vy;
  if (*ix < *iy) return -1;
  if (*ix > *iy) return  1;
  return 0;
}

static int* compare_indirect_map;
static int compare_indirect_ints(void* vx, void* vy)
{
  int* ix = (int*)vx;
  int* iy = (int*)vy;
  if (compare_indirect_map[*ix] < compare_indirect_map[*iy]) return -1;
  if (compare_indirect_map[*ix] > compare_indirect_map[*iy]) return  1;
  return 0;
}

/*************************************************************/
/* radix sort                                                */
/*************************************************************/

#if 0
/* NCOUNTS = 2^LOGRADIX */

#define RADIX_SORT_LOGRADIX 4
#define RADIX_SORT_NCOUNTS  16

static unsigned int counts[RADIX_SORT_NCOUNTS];

static int
radix_sort(unsigned int* x, int n)
{
  int i;
  unsigned int mask;

  unsigned int  ncounts;

  unsigned int* y;
  unsigned int* to;
  unsigned int* from;

  unsigned int v;
  unsigned int partialsum;
  unsigned int next;
  unsigned int bits_sorted;

  if (RADIX_SORT_LOGRADIX >= 8*sizeof(unsigned int)) {
    taucs_printf("radix sort: radix too large.\n");
    /* the computation of ncounts will fail */
    return 0;
  }

  mask    = 0;
  ncounts = 1;
  for (i=0; i<RADIX_SORT_LOGRADIX; i++) {
    mask = (mask << 1) | 1;
    ncounts = ncounts << 1;
  }

  assert(ncounts==RADIX_SORT_NCOUNTS);

  y      = (unsigned int*) malloc(n       * sizeof(unsigned int));
  if (!y) {
    taucs_printf("radix sort: out of memory.\n");
    return -1;
  }

  from = x;
  to   = y;

  bits_sorted = 0;
  while(bits_sorted < 8*sizeof(unsigned int)) {
    for (i=0; i<ncounts; i++) counts[i] = 0;

    for (i=0; i<n; i++) {
      v = (from[i] >> bits_sorted) & mask;
      assert(v < ncounts);
      counts[v] ++;
    }

    partialsum = 0;
    for (i=0; i<ncounts; i++) {
      /*printf("<%d ",counts[i]);*/
      next = counts[i];
      counts[i] = partialsum;
      /*printf("%d>\n",counts[i]);*/
      partialsum = partialsum + next;
    }

    for (i=0; i<n; i++) {
      v = (from[i] >> bits_sorted) & mask;
      assert(counts[v] < n);
      to[counts[v]] = from[i];
      counts[v] ++;
    }
    /*
    printf("===========\n");
    for (i=0; i<n; i++) printf(">>%d>> %08x\n",bits_sorted,to[i]);
    printf("===========\n");
    */

    bits_sorted += RADIX_SORT_LOGRADIX;
    if (from == x) {
      from = y;
      to   = x;
    } else {
      from = x;
      to   = y;
    } 
  }

  if (from == y) 
    for (i=0; i<n; i++) x[i] = y[i];

  free(y);

  return 0;
}
#endif

/*************************************************************/
/* create and free the factor object                         */
/*************************************************************/

static supernodal_factor_matrix*
multifrontal_supernodal_create()
{
  supernodal_factor_matrix* L;
  
  L = (supernodal_factor_matrix*) malloc(sizeof(supernodal_factor_matrix));
  if (!L) return NULL;

#ifdef TAUCS_CORE_SINGLE
  L->flags = TAUCS_SINGLE;
#endif

#ifdef TAUCS_CORE_DOUBLE
  L->flags = TAUCS_DOUBLE;
#endif

#ifdef TAUCS_CORE_SCOMPLEX
  L->flags = TAUCS_SCOMPLEX;
#endif

#ifdef TAUCS_CORE_DCOMPLEX
  L->flags = TAUCS_DCOMPLEX;
#endif

  L->uplo      = 'l';
  L->n         = -1; /* unused */

  L->sn_struct   = NULL;
  L->sn_size     = NULL;
  L->sn_up_size  = NULL;
  L->parent      = NULL;
  L->first_child = NULL;
  L->next_child  = NULL;
  L->sn_blocks_ld  = NULL;
  L->sn_blocks     = NULL;
  L->up_blocks_ld  = NULL;
  L->up_blocks     = NULL;

  return L;
}

void taucs_dtl(supernodal_factor_free)(void* vL)
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  int sn;

  if (!L) return;
  
  free(L->parent);
  free(L->first_child);
  free(L->next_child);

  free(L->sn_size);
  free(L->sn_up_size);
  free(L->sn_blocks_ld);
  free(L->up_blocks_ld);

  if (L->sn_struct)   
    for (sn=0; sn<L->n_sn; sn++)
      free(L->sn_struct[sn]);

  if (L->sn_blocks)   
    for (sn=0; sn<L->n_sn; sn++)
      free(L->sn_blocks[sn]);

  if (L->up_blocks)   
    for (sn=0; sn<L->n_sn; sn++)
      free(L->up_blocks[sn]);

  free(L->sn_struct);
  free(L->sn_blocks);
  free(L->up_blocks);

  free(L);
}

void taucs_dtl(supernodal_factor_free_numeric)(void* vL)
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  int sn;
  
  for (sn=0; sn<L->n_sn; sn++) {
    free(L->sn_blocks[sn]);
    L->sn_blocks[sn] = NULL;
    free(L->up_blocks[sn]);
    L->up_blocks[sn] = NULL;
  }
}

taucs_ccs_matrix*
taucs_dtl(supernodal_factor_to_ccs)(void* vL)
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  taucs_ccs_matrix* C;
  int n,nnz;
  int i,j,ip,jp,sn,next;
  taucs_datatype v;
  int* len;

  n = L->n;

  len = (int*) malloc(n*sizeof(int));
  if (!len) return NULL;

  nnz = 0;
  /*
  for (sn=0; sn<L->n_sn; sn++) {
    for (jp=0; jp<(L->sn_size)[sn]; jp++) {
      j = (L->sn_struct)[sn][jp];
      len[j] = (L->sn_up_size)[sn] - jp;
      nnz += len[j];
    }
  }
  */

  for (sn=0; sn<L->n_sn; sn++) {
    for (jp=0; jp<(L->sn_size)[sn]; jp++) {
      j = (L->sn_struct)[sn][jp];
      len[j] = 0;

      for (ip=jp; ip<(L->sn_size)[sn]; ip++) {
	i = (L->sn_struct)[sn][ ip ];
	v = (L->sn_blocks)[sn][ jp*(L->sn_blocks_ld)[sn] + ip ];

	if (taucs_re(v) || taucs_im(v)) { 
	  len[j] ++;
	  nnz ++;
	}
      }
      for (ip=(L->sn_size)[sn]; ip<(L->sn_up_size)[sn]; ip++) {
	i = (L->sn_struct)[sn][ ip ];
	v = (L->up_blocks)[sn][ jp*(L->up_blocks_ld)[sn] + (ip-(L->sn_size)[sn]) ];

	if (taucs_re(v) || taucs_im(v)) { 
	  len[j] ++;
	  nnz ++;
	}
      }
    }
  }


  C = taucs_dtl(ccs_create)(n,n,nnz);
  if (!C) {
    free(len);
    return NULL;
  }

#ifdef TAUCS_CORE_SINGLE
  C->flags = TAUCS_SINGLE;
#endif

#ifdef TAUCS_CORE_DOUBLE
  C->flags = TAUCS_DOUBLE;
#endif

#ifdef TAUCS_CORE_SCOMPLEX
  C->flags = TAUCS_SCOMPLEX;
#endif

#ifdef TAUCS_CORE_DCOMPLEX
  C->flags = TAUCS_DCOMPLEX;
#endif

  C->flags = TAUCS_TRIANGULAR | TAUCS_LOWER;



  (C->colptr)[0] = 0;
  for (j=1; j<=n; j++) (C->colptr)[j] = (C->colptr)[j-1] + len[j-1];

  free(len);

  for (sn=0; sn<L->n_sn; sn++) {
    for (jp=0; jp<(L->sn_size)[sn]; jp++) {
      j = (L->sn_struct)[sn][jp];

      next = (C->colptr)[j];

      /*
      memcpy((C->rowind) + next,
	     ((L->sn_struct)[sn]) + jp,
	     ((L->sn_up_size)[sn] - jp) * sizeof(int));
      memcpy((C->taucs_values) + next,
	     ((L->sn_blocks)[sn]) + (jp*(L->sn_blocks_ld)[sn] + jp),
	     ((L->sn_size)[sn] - jp) * sizeof(taucs_datatype));
      next += ((L->sn_size)[sn] - jp);
      memcpy((C->taucs_values) + next,
	     ((L->up_blocks)[sn]) + jp*(L->up_blocks_ld)[sn],
	     ((L->sn_up_size)[sn] - (L->sn_size)[sn]) * sizeof(taucs_datatype));
      */

      for (ip=jp; ip<(L->sn_size)[sn]; ip++) {
	i = (L->sn_struct)[sn][ ip ];
	v = (L->sn_blocks)[sn][ jp*(L->sn_blocks_ld)[sn] + ip ];

	if (!taucs_re(v) && !taucs_im(v)) continue;
	/*if (v == 0.0) continue;*/

	(C->rowind)[next] = i;
	(C->taucs_values)[next] = v;
	next++;
      }
      for (ip=(L->sn_size)[sn]; ip<(L->sn_up_size)[sn]; ip++) {
	i = (L->sn_struct)[sn][ ip ];
	v = (L->up_blocks)[sn][ jp*(L->up_blocks_ld)[sn] + (ip-(L->sn_size)[sn]) ];

	if (!taucs_re(v) && !taucs_im(v)) continue;
	/*if (v == 0.0) continue;*/

	(C->rowind)[next] = i;
	(C->taucs_values)[next] = v;
	next++;
      }
    }
  }

  return C;
}

/*************************************************************/
/* create and free frontal matrices                          */
/*************************************************************/

static supernodal_frontal_matrix* 
supernodal_frontal_create(int* firstcol_in_supernode,
			  int sn_size,
			  int n, 
			  int* rowind)
{
  supernodal_frontal_matrix* tmp;

  tmp = (supernodal_frontal_matrix*)malloc(sizeof(supernodal_frontal_matrix));
  if(tmp==NULL) return NULL;

  tmp->sn_size = sn_size;
  tmp->n = n;

  tmp->rowind = rowind;

  tmp->n = n;
  tmp->sn_size = sn_size;
  tmp->up_size = n-sn_size;

  tmp->sn_vertices = rowind;
  tmp->up_vertices = rowind + sn_size;

  tmp->SFM_F1 = (taucs_datatype*)calloc((tmp->sn_size)*(tmp->sn_size),sizeof(taucs_datatype));
  tmp->SFM_F2 = (taucs_datatype*)calloc((tmp->up_size)*(tmp->sn_size),sizeof(taucs_datatype));
  tmp->SFM_U  = (taucs_datatype*)calloc((tmp->up_size)*(tmp->up_size),sizeof(taucs_datatype));

  if(tmp->SFM_F1 == NULL || tmp->SFM_F2==NULL || tmp->SFM_U==NULL) {
    free(tmp->SFM_U);
    free(tmp->SFM_F1);
    free(tmp->SFM_F2);
    free(tmp);
    return NULL;
  }

  assert(tmp);
  return tmp;
}

static void supernodal_frontal_free(supernodal_frontal_matrix* to_del)
{
  /* 
     SFM_F1 and SFM_F2 are moved to the factor,
     but this function may be called before they are
     moved.
  */


  if (to_del) {
    free(to_del->SFM_F1);
    free(to_del->SFM_F2);
    free(to_del->SFM_U);
    free(to_del);
  }
}

/*************************************************************/
/* factor a frontal matrix                                   */
/*************************************************************/

static int
multifrontal_supernodal_front_factor(int sn,
				     int* firstcol_in_supernode,
				     int sn_size,
				     taucs_ccs_matrix* A,
				     supernodal_frontal_matrix* mtr,
				     int* bitmap,
				     supernodal_factor_matrix* snL)
{
  int i,j;
  int* ind;
  taucs_datatype* re;
  int INFO;

  /* creating transform for real indices */
  for(i=0;i<mtr->sn_size;i++) bitmap[mtr->sn_vertices[i]] = i;
  for(i=0;i<mtr->up_size;i++) bitmap[mtr->up_vertices[i]] = mtr->sn_size + i;

  /* adding sn_size column of A to first sn_size column of frontal matrix */

  for(j=0;j<(mtr->sn_size);j++) {
    ind = &(A->rowind[A->colptr[*(firstcol_in_supernode+j)]]);
    re  = &(A->taucs_values[A->colptr[*(firstcol_in_supernode+j)]]); 
    for(i=0;
	i < A->colptr[*(firstcol_in_supernode+j)+1] 
            - A->colptr[*(firstcol_in_supernode+j)];
	i++) {
      if (bitmap[ind[i]] < mtr->sn_size)
	mtr->SFM_F1[ (mtr->sn_size)*j + bitmap[ind[i]]] =
	  taucs_add( mtr->SFM_F1[ (mtr->sn_size)*j + bitmap[ind[i]]] , re[i] );
      else
	mtr->SFM_F2[ (mtr->up_size)*j + bitmap[ind[i]] - mtr->sn_size] =
	  taucs_add( mtr->SFM_F2[ (mtr->up_size)*j + bitmap[ind[i]] - mtr->sn_size] , re[i] );
    }
  }

  /* we use the BLAS through the Fortran interface */

  /* solving of lower triangular system for L */
  if (mtr->sn_size)
    taucs_potrf ("LOWER",
		 &(mtr->sn_size),
		 mtr->SFM_F1,&(mtr->sn_size),
		 &INFO);


  if (INFO) {
    taucs_printf("\t\tLL^T Factorization: Matrix is not positive definite.\n");
    taucs_printf("\t\t                    nonpositive pivot in column %d\n",
		 mtr->sn_vertices[INFO-1]);
    return -1;
  }

  /* getting completion for found columns of L */
  if (mtr->up_size && mtr->sn_size)
    taucs_trsm ("Right",
		"Lower",
		"Conjugate",
		"No unit diagonal",
		&(mtr->up_size),&(mtr->sn_size),
		&taucs_one_const,
		mtr->SFM_F1,&(mtr->sn_size),
		mtr->SFM_F2,&(mtr->up_size));

  (snL->sn_blocks   )[sn] = mtr->SFM_F1;
  (snL->sn_blocks_ld)[sn] = mtr->sn_size;

  (snL->up_blocks   )[sn] = mtr->SFM_F2;
  (snL->up_blocks_ld)[sn] = mtr->up_size;
  /* printf("*** sn=%d up_ld=%d (%d)\n",sn,mtr->up_size,(snL->up_vertex_ptr)[sn+1] - (snL->up_vertex_ptr)[sn]);*/

  /* computation of updated part of frontal matrix */
  if (mtr->up_size && mtr->sn_size)
    taucs_herk ("Lower",
		"No Conjugate",
		&(mtr->up_size),&(mtr->sn_size),
		&taucs_minusone_real_const,
		mtr->SFM_F2,&(mtr->up_size),
		&taucs_one_real_const,
		mtr->SFM_U, &(mtr->up_size));

  mtr->SFM_F1 = NULL; /* so we don't free twice */
  mtr->SFM_F2 = NULL; /* so we don't free twice */

  return 0;
 }

/*************************************************************/
/* extend-add                                                */
/*************************************************************/

static void 
multifrontal_supernodal_front_extend_add(
					 supernodal_frontal_matrix* parent_mtr,
					 supernodal_frontal_matrix* my_mtr,
					 int* bitmap)
{
  int j,i,parent_i,parent_j;
  taucs_datatype v;

  for(i=0;i<parent_mtr->sn_size;i++) bitmap[parent_mtr->sn_vertices[i]] = i;
  for(i=0;i<parent_mtr->up_size;i++) bitmap[parent_mtr->up_vertices[i]] = (parent_mtr->sn_size)+i;

  /* extend add operation for update matrix */
  for(j=0;j<my_mtr->up_size;j++) {
    for(i=j;i<my_mtr->up_size;i++) {
      parent_j = bitmap[ my_mtr->up_vertices[j] ];
      parent_i = bitmap[ my_mtr->up_vertices[i] ];
      /* we could skip this if indices were sorted */
      if (parent_j>parent_i) {
	int tmp = parent_j;
	parent_j = parent_i;
	parent_i = tmp;
      }

      v = (my_mtr->SFM_U)[(my_mtr->up_size)*j+i];

      if (parent_j < parent_mtr->sn_size) {
	if (parent_i < parent_mtr->sn_size) {
	  (parent_mtr->SFM_F1)[ (parent_mtr->sn_size)*parent_j + parent_i] =
	    taucs_add( (parent_mtr->SFM_F1)[ (parent_mtr->sn_size)*parent_j + parent_i] , v );
	} else {
	  (parent_mtr->SFM_F2)[ (parent_mtr->up_size)*parent_j + (parent_i-parent_mtr->sn_size)] =
	    taucs_add( (parent_mtr->SFM_F2)[ (parent_mtr->up_size)*parent_j + (parent_i-parent_mtr->sn_size)] , v );
	}
      } else {
	(parent_mtr->SFM_U)[ (parent_mtr->up_size)*(parent_j-parent_mtr->sn_size) + (parent_i-parent_mtr->sn_size)] =
	  taucs_add( (parent_mtr->SFM_U)[ (parent_mtr->up_size)*(parent_j-parent_mtr->sn_size) + (parent_i-parent_mtr->sn_size)] , v);
      }
    }
  }
}

/*************************************************************/
/* symbolic elimination                                      */
/*************************************************************/

/* UNION FIND ROUTINES */

static int uf_makeset(int* uf, int i)        { uf[i] = i; return i; }
static int uf_find   (int* uf, int i)        { if (uf[i] != i) 
                                                 uf[i] = uf_find(uf,uf[i]); 
                                               return uf[i]; }
static int uf_union  (int* uf, int s, int t) {
  /*assert(uf_find(uf,s) <= uf_find(uf,t));*/
  
  if (uf_find(uf,s) < uf_find(uf,t)) {
    uf[uf_find(uf,s)] = uf_find(uf,t); 
    return (uf_find(uf,t)); 
  } else {
    uf[uf_find(uf,s)] = uf_find(uf,t); 
    return (uf_find(uf,t)); 
  }
}

static
void recursive_postorder(int  j,
			 int  first_child[],
			 int  next_child[],
			 int  postorder[],
			 int  ipostorder[],
			 int* next)
{
  int c;
  for (c=first_child[j]; c != -1; c = next_child[c]) {
    /*printf("*** %d is child of %d\n",c,j);*/
    recursive_postorder(c,first_child,next_child,
			postorder,ipostorder,next);
  }
  /*printf(">>> j=%d next=%d\n",j,*next);*/
  if (postorder)  postorder [*next] = j;
  if (ipostorder) ipostorder[j] = *next;
  (*next)++;
}

#define GILBERT_NG_PEYTON_ANALYSIS_SUP

/* in a few tests the supernodal version seemed slower */
#undef GILBERT_NG_PEYTON_ANALYSIS_SUP

static int ordered_uf_makeset(int* uf, int i)
{ 
  uf[i] = i; 
  return i; 
}
static int ordered_uf_find   (int* uf, int i) 
{ 
  if (uf[i] != i) 
    uf[i] = uf_find(uf,uf[i]); 
  return uf[i]; 
}
static int ordered_uf_union  (int* uf, int s, int t) 
{
  assert(uf[t] == t);
  assert(uf[s] == s);
  assert(t > s);
  if (t > s) {
    uf[s] = t; 
    return t; 
  } else
    uf[t] = s; 
    return s; 
}

static void 
tree_level(int j,
	   int isroot, 
	   int first_child[],
	   int next_child[],
	   int level[],
	   int level_j)
{
  int c;
  if (!isroot) level[j] = level_j;
  for (c=first_child[j]; c != -1; c = next_child[c]) {
    tree_level(c,
	       FALSE,
	       first_child,
	       next_child,
	       level,
	       level_j+1);
  }
}

static void
tree_first_descendant(int j,
		      int isroot, 
		      int first_child[],
		      int next_child[],
		      int ipostorder[],
		      int first_descendant[])
{
  int c;
  int fd = ipostorder[j];
  for (c=first_child[j]; c != -1; c = next_child[c]) {
    tree_first_descendant(c,
			  FALSE,
			  first_child,
			  next_child,
			  ipostorder,
			  first_descendant);
    if (first_descendant[c] < fd) fd = first_descendant[c]; 
  }
  if (!isroot) first_descendant[j] = fd;
}


int
taucs_ccs_etree(taucs_ccs_matrix* A,
		int* parent,
		int* l_colcount,
		int* l_rowcount,
		int* l_nnz);

int 
taucs_ccs_etree_liu(taucs_ccs_matrix* A,
		    int* parent,
		    int* l_colcount,
		    int* l_rowcount,
		    int* l_nnz);

#ifdef TAUCS_CORE_DOUBLE
int
taucs_ccs_etree(taucs_ccs_matrix* A,
		int* parent,
		int* l_colcount,
		int* l_rowcount,
		int* l_nnz)
{
  int* prev_p;
  int* prev_nbr;
  int* level;
  int* first_descendant;
  int* l_cc;
  int* l_rc;
  int* wt;

  int  n = A->n;
  int  u,p,q,pprime;
  int  ju;
  int* postorder;
  int* ipostorder;
  int  *first_child,*next_child;

  int i,j,k,ip,jp,kp;
  int nnz,jnnz;
  int* uf;
  int* rowptr;
  int* colind;
  int* rowcount;
  int* realroot;

  /* we need the row structures for the lower triangle */

  nnz = (A->colptr)[n];
  
  uf       = malloc(n     * sizeof(int));
  rowcount = malloc((n+1) * sizeof(int));
  rowptr   = malloc((n+1) * sizeof(int));
  colind   = malloc(nnz   * sizeof(int));

  if (!uf || !rowcount || !rowptr || !colind) {
    free(uf);
    free(rowcount);
    free(rowptr);
    free(colind);
    return -1;
  }

  for (i=0; i <=n; i++) rowcount[i] = 0;
  for (j=0; j < n; j++) {
    jnnz = (A->colptr)[j+1] - (A->colptr)[j];
    for (ip=0; ip<jnnz; ip++) {
      i = (A->rowind)[(A->colptr)[j] + ip];
      if (j < i) rowcount[i]++;
    }
  }

  ip = 0;
  for (i=0; i <= n; i++) {
    int next_ip = ip + rowcount[i];
    rowcount[i] = ip;
    rowptr  [i] = ip;
    ip = next_ip;
  }

  for (j=0; j < n; j++) {
    jnnz = (A->colptr)[j+1] - (A->colptr)[j];
    for (ip=0; ip<jnnz; ip++) {
      i = (A->rowind)[(A->colptr)[j] + ip];
      if (i==j) continue;
      assert( rowcount[i] < rowptr[i+1] );
      colind[ rowcount[i] ] = j;
      rowcount[i]++;
    }
  }

  /* now compute the etree */

  {
    int u,t,vroot;
    realroot = rowcount; /* reuse space */

    for (i=0; i<n; i++) {
      uf_makeset(uf,i);
      realroot[i] = i;
      parent[i] = n;
      vroot = i;
      for (kp=rowptr[i]; kp<rowptr[i+1]; kp++) {
	k=colind[kp];
	u = uf_find(uf,k);
	t = realroot[u];
	if (parent[t] == n && t != i) {
	  parent[t] = i;
	  vroot = uf_union(uf,vroot,u);
	  realroot[vroot] = i;
	}
      }
    }
  }

  free(colind);
  free(rowptr);
  free(rowcount);

  /* now only uf remains allocated */
  
  /* compute column counts */

  if (l_colcount || l_rowcount || l_nnz) {
    int* l_nz;
    int  tmp;
    int  u,p,q;

    first_child = malloc((n+1) * sizeof(int));
    next_child  = malloc((n+1) * sizeof(int));
    postorder   = malloc(n     * sizeof(int));
    ipostorder  = malloc(n     * sizeof(int));
    wt          = malloc(n     * sizeof(int));
    level       = malloc(n     * sizeof(int));
    prev_p      = malloc(n     * sizeof(int));

#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
    prev_nbr         = malloc(n     * sizeof(int));
    first_descendant = malloc(n     * sizeof(int));
#endif

    /* we allocate scratch vectors to avoid conditionals */
    /* in the inner loop.                                */

    if (l_colcount) l_cc = l_colcount;
    else            l_cc = (int*) malloc(n*sizeof(int));
    if (l_rowcount) l_rc = l_rowcount;
    else            l_rc = (int*) malloc(n*sizeof(int));
    if (l_nnz)      l_nz = l_nnz;
    else            l_nz = &tmp;


    if (!first_child || !next_child || !postorder
	|| !ipostorder || !wt || !level || !prev_p
	|| (!l_colcount && !l_cc) 
	|| (!l_rowcount && !l_rc) 
#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
	|| !prev_nbr || !first_descendant
#endif
	) {
      free(uf);

      if (!l_colcount) free(l_cc);
      if (!l_rowcount) free(l_rc);

      free(postorder);
      free(ipostorder);
      free(wt);
      free(level);
      free(prev_p);
      
#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
      free(prev_nbr);
      free(first_descendant);
#endif
      return -1;
    }

    /*for (j=0; j<n; j++) printf("parent[%d] = %d\n",j,parent[j]);*/

    /* compute the postorder */
    
    for (j=0; j<=n; j++) first_child[j] = -1;
    for (j=n-1; j>=0; j--) {
      next_child[j] = first_child[parent[j]];
      first_child[parent[j]] = j;
    }
    
    {
      int next = 0;
      recursive_postorder(n,first_child,next_child,
			  postorder,
			  ipostorder,&next);
    }
    
    /* sort by postorder of etree */
    /* compute level, fst_desc */

    tree_level(n,TRUE,first_child,next_child,
	       level,-1);
    
    for (u=0; u < n; u++) prev_p  [u] = -1;
    for (u=0; u < n; u++) l_rc    [u] =  1;
    for (u=0; u < n; u++) ordered_uf_makeset(uf,u);
    for (u=0; u < n; u++) {
      if (first_child[u] == -1)
	wt[u] = 1; /* leaves     */
      else
	wt[u] =  0; /* nonleaves */
    }

#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
    for (u=0; u < n; u++) prev_nbr[u] = -1;

    tree_first_descendant(n,TRUE,
			  first_child,next_child,ipostorder,
			  first_descendant);
#endif

    free(first_child);
    free(next_child);

    for (p=0; p<n; p++) {
      jp = postorder[p];
      if (parent[jp] != n) wt[parent[jp]] --;
      for (ip = (A->colptr)[jp]; ip < (A->colptr)[jp+1]; ip++) {
	ju = (A->rowind)[ip];
	u  = ipostorder[ju];
	if (ju==jp) continue; /* we only want proper neighbors */
#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
	if (first_descendant[jp] > prev_nbr[u]) {
#else
	if (1) {
#endif
	  wt[jp] ++;
	  pprime = prev_p[ju];
	  if (pprime == -1) 
	    l_rc[ju] += level[jp] - level[ju];
	  else {
	    q = ordered_uf_find(uf,pprime);
	    l_rc[ju] += level[jp] - level[q];
	    wt[q] --;
	  }
	  prev_p[ju] = jp;
	}
#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
	prev_nbr[u] = p;
#endif
      }
      if (parent[jp] != n) {
	if (!(ipostorder[parent[jp]] > ipostorder[jp])) {
	  printf("jp %d parent %d (ipo_j %d ipo_parent %d\n",
		 jp,parent[jp],ipostorder[jp],ipostorder[parent[jp]]);
	}
	assert(ipostorder[parent[jp]] > ipostorder[jp]);
	ordered_uf_union(uf,jp,parent[jp]);
      }
    }

    *l_nz = 0;
    for (u=0; u<n; u++) {
      l_cc[u] = wt[u];
      *l_nz += wt[u];
    }
    for (u=0; u<n; u++) {
      if (parent[u] != n) {
	l_cc[parent[u]] += l_cc[u];
	*l_nz += l_cc[u];
      }
    }

    /* free scrtach vectors                              */

    if (!l_colcount) free(l_cc);
    if (!l_rowcount) free(l_rc);

    /* free other data structures */

    free(postorder);
    free(ipostorder);
    free(wt);
    free(level);
    free(prev_p);
    
#ifdef GILBERT_NG_PEYTON_ANALYSIS_SUP
    free(prev_nbr);
    free(first_descendant);
#endif
  }

  free(uf);

  return 0;
}

int 
taucs_ccs_etree_liu(taucs_ccs_matrix* A,
		    int* parent,
		    int* l_colcount,
		    int* l_rowcount,
		    int* l_nnz)
{
  int n = A->n;
  int i,j,k,ip,jp,kp;
  int nnz,jnnz;

  int* uf;
  int* rowptr;
  int* colind;

  int* rowcount;
  int* marker;
  int* realroot;

  int* l_cc;
  int* l_rc;

  /* we need the row structures for the lower triangle */

  nnz = (A->colptr)[n];
  
  uf       = malloc(n     * sizeof(int));
  rowcount = malloc((n+1) * sizeof(int));
  rowptr   = malloc((n+1) * sizeof(int));
  colind   = malloc(nnz   * sizeof(int));

  for (i=0; i <=n; i++) rowcount[i] = 0;

  for (j=0; j < n; j++) {
    
    jnnz = (A->colptr)[j+1] - (A->colptr)[j];

    for (ip=0; ip<jnnz; ip++) {
      i = (A->rowind)[(A->colptr)[j] + ip];
      if (j < i) rowcount[i]++;
    }
  }

  ip = 0;
  for (i=0; i <= n; i++) {
    int next_ip = ip + rowcount[i];
    rowcount[i] = ip;
    rowptr  [i] = ip;
    ip = next_ip;
  }

  for (j=0; j < n; j++) {
    jnnz = (A->colptr)[j+1] - (A->colptr)[j];

    for (ip=0; ip<jnnz; ip++) {
      i = (A->rowind)[(A->colptr)[j] + ip];
      if (i==j) continue;
      assert( rowcount[i] < rowptr[i+1] );
      colind[ rowcount[i] ] = j;
      rowcount[i]++;
    }
  }

  /* now compute the etree */

  {
    int u,t,vroot;
    realroot = rowcount; /* reuse space */

    for (i=0; i<n; i++) {
      uf_makeset(uf,i);
      realroot[i] = i;
      parent[i] = n;
      vroot = i;
      for (kp=rowptr[i]; kp<rowptr[i+1]; kp++) {
	k=colind[kp];
	u = uf_find(uf,k);
	t = realroot[u];
	if (parent[t] == n && t != i) {
	  parent[t] = i;
	  vroot = uf_union(uf,vroot,u);
	  realroot[vroot] = i;
	}
      }
    }
  }

  /* compute column counts */

  if (l_colcount || l_rowcount || l_nnz) {
    int* l_nz;
    int  tmp;

    /* we allocate scratch vectors to avoid conditionals */
    /* in the inner loop.                                */

    if (l_colcount) l_cc = l_colcount;
    else            l_cc = (int*) malloc(n*sizeof(int));
    if (l_rowcount) l_rc = l_rowcount;
    else            l_rc = (int*) malloc(n*sizeof(int));
    if (l_nnz)      l_nz = l_nnz;
    else            l_nz = &tmp;

    marker = rowcount; /* we reuse the space */
    
    for (j=0; j < n; j++) l_cc[j] = 1;
    *l_nz = n;
    
    for (i=0; i<n; i++) marker[i] = n; /* clear the array */
    
    for (i=0; i<n; i++) {
      l_rc[i] = 1;
      marker[ i ] = i;
      for (kp=rowptr[i]; kp<rowptr[i+1]; kp++) {
	k=colind[kp];
	j=k;
	while (marker[j] != i) {
	  l_cc[j]++;
	  l_rc[i]++;
	  (*l_nz)++;
	  marker[j] = i;
	  j = parent[j];
	}
      }
    }

    /* free scrtach vectors                              */

    if (!l_colcount) free(l_cc);
    if (!l_rowcount) free(l_rc);
  }

  free(colind);
  free(rowptr);
  free(rowcount);
  free(uf);

  return 0;
}
#endif

static int
recursive_symbolic_elimination(int            j,
			       taucs_ccs_matrix* A,
			       int            first_child[],
			       int            next_child[],
			       int*           n_sn,
			       int            sn_size[],
			       int            sn_up_size[],
			       int*           sn_rowind[],
			       int            sn_first_child[], 
			       int            sn_next_child[], 
			       int            rowind[],
			       int            column_to_sn_map[],
			       int            map[],
			       int            do_order,
			       int            ipostorder[]
			       )
{
  int  i,ip,c,c_sn;
  int  in_previous_sn;
  int  nnz;
  
  for (c=first_child[j]; c != -1; c = next_child[c]) {
    if (recursive_symbolic_elimination(c,A,
				       first_child,next_child,
				       n_sn,
				       sn_size,sn_up_size,sn_rowind,
				       sn_first_child,sn_next_child,
				       rowind, /* temporary */
				       column_to_sn_map,
				       map,
				       do_order,ipostorder
				       ) 
	== -1) return -1;
  }

  in_previous_sn = 1;
  if (j == A->n) 
    in_previous_sn = 0; /* this is not a real column */
  else if (first_child[j] == -1) 
    in_previous_sn = 0; /* this is a leaf */
  else if (next_child[first_child[j]] != -1) 
    in_previous_sn = 0; /* more than 1 child */
  else { 
    /* check that the structure is nested */
    /* map contains child markers         */

    c=first_child[j];
    for (ip=(A->colptr)[j]; ip<(A->colptr)[j+1]; ip++) {
      i = (A->rowind)[ip];
      in_previous_sn = in_previous_sn && (map[i] == c);
    }
  }

  if (in_previous_sn) {
    c = first_child[j];
    c_sn = column_to_sn_map[c];
    column_to_sn_map[j] = c_sn;

    /* swap row indices so j is at the end of the */
    /* supernode, not in the update indices       */
    for (ip=sn_size[c_sn]; ip<sn_up_size[c_sn]; ip++) 
      if (sn_rowind[c_sn][ip] == j) break;
    assert(ip<sn_up_size[c_sn]);
    sn_rowind[c_sn][ip] = sn_rowind[c_sn][sn_size[c_sn]];
    sn_rowind[c_sn][sn_size[c_sn]] = j;

    /* mark the nonzeros in the map */
    for (ip=sn_size[c_sn]; ip<sn_up_size[c_sn]; ip++) 
      map[ sn_rowind[c_sn][ip] ] = j;

    sn_size   [c_sn]++;

    return 0;
  }

  /* we are in a new supernode */

  if (j < A->n) {
    nnz = 1;
    rowind[0] = j;
    map[j]    = j;
    
    for (c=first_child[j]; c != -1; c = next_child[c]) {
      c_sn = column_to_sn_map[c];
      for (ip=sn_size[c_sn]; ip<sn_up_size[c_sn]; ip++) {
	i = sn_rowind[c_sn][ip];
	if (i > j && map[i] != j) { /* new row index */
	  map[i] = j;
	  rowind[nnz] = i;
	  nnz++;
	}
      }
    }
    
    for (ip=(A->colptr)[j]; ip<(A->colptr)[j+1]; ip++) {
      i = (A->rowind)[ip];
      if (map[i] != j) { /* new row index */
	map[i] = j;
	rowind[nnz] = i;
	nnz++;
      }
    }
  }
    
  /*printf("children of sn %d: ",*n_sn);*/
  for (c=first_child[j]; c != -1; c = next_child[c]) {
    c_sn = column_to_sn_map[c];
    /*printf("%d ",c_sn);*/
    if (c==first_child[j])
      sn_first_child[*n_sn] = c_sn;
    else {
      sn_next_child[ c_sn ] = sn_first_child[*n_sn];
      sn_first_child[*n_sn] = c_sn;
    }
  }
  /*printf("\n");*/

  if (j < A->n) {
    column_to_sn_map[j] = *n_sn;
    sn_size   [*n_sn] = 1;
    sn_up_size[*n_sn] = nnz;
    sn_rowind [*n_sn] = (int*) malloc(nnz * sizeof(int));
    if (!( sn_rowind [*n_sn] )) return -1;
    for (ip=0; ip<nnz; ip++) sn_rowind[*n_sn][ip] = rowind[ip];
    if (do_order) {
      /* Sivan and Vladimir: we think that we can sort in */
      /* column order, not only in etree postorder.       */
      /*
	radix_sort(sn_rowind [*n_sn],nnz);
	qsort(sn_rowind [*n_sn],nnz,sizeof(int),compare_ints);
      */
      compare_indirect_map = ipostorder;
      qsort(sn_rowind [*n_sn],nnz,sizeof(int),compare_indirect_ints);
    }
    assert(sn_rowind [*n_sn][0] == j);
    (*n_sn)++;
  }

  return 0;
}

/* count zeros and nonzeros in a supernode to compute the */
/* utility of merging fundamental supernodes.             */

typedef struct {
  double zeros;
  double nonzeros;
} znz;

static znz
recursive_amalgamate_supernodes(int           sn,
				int*           n_sn,
				int            sn_size[],
				int            sn_up_size[],
				int*           sn_rowind[],
				int            sn_first_child[], 
				int            sn_next_child[], 
				int            rowind[],
				int            column_to_sn_map[],
				int            map[],
				int            do_order,
				int            ipostorder[]
				)
{
  int  i,ip,c_sn,gc_sn;
  /*int  i,ip,c,c_sn,gc_sn;*/
  int  nnz;
  int  nchildren /*, ichild*/; /* number of children, child index */
  znz* c_znz;
  znz  sn_znz, merged_znz;
  /*int zero_count = 0;*/
  int new_sn_size, new_sn_up_size;

  sn_znz.zeros    = 0.0;
  sn_znz.nonzeros = (double) (((sn_up_size[sn] - sn_size[sn]) * sn_size[sn]) 
                              + (sn_size[sn] * (sn_size[sn] + 1))/2);

  if (sn_first_child[sn] == -1) { /* leaf */
    return sn_znz;
  }

  nchildren = 0;
  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn])
    nchildren++;

  c_znz = (znz*) alloca(nchildren * sizeof(znz));

  /*printf("supernode %d out of %d\n",sn,*n_sn);*/

  /* merge the supernode with its children! */

  i = 0;
  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
    c_znz[i] = 
      recursive_amalgamate_supernodes(c_sn,
				      n_sn,
				      sn_size,sn_up_size,sn_rowind,
				      sn_first_child,sn_next_child,
				      rowind, /* temporary */
				      column_to_sn_map,
				      map,
				      do_order,ipostorder
				      );
    assert(c_znz[i].zeros + c_znz[i].nonzeros ==
	   (double) (((sn_up_size[c_sn] - sn_size[c_sn]) * sn_size[c_sn]) 
		     + (sn_size[c_sn] * (sn_size[c_sn] + 1))/2 ));
    i++;
  }

  merged_znz.nonzeros = sn_znz.nonzeros;
  merged_znz.zeros    = sn_znz.zeros;
                   
  for (i=0; i<nchildren; i++) {
    merged_znz.nonzeros += (c_znz[i]).nonzeros;
    merged_znz.zeros    += (c_znz[i]).zeros;
  }

  /*  printf("supernode %d out of %d (continuing)\n",sn,*n_sn);*/

  /* should we merge the supernode with its children? */

  nnz = 0;
  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
    for (ip=0; ip<sn_size[c_sn]; ip++) {
      i = sn_rowind[c_sn][ip];
      assert( map[i] != sn );
      map[i] = sn;
      rowind[nnz] = i;
      nnz++;
    }
  }

  for (ip=0; ip<sn_size[sn]; ip++) {
    i = sn_rowind[sn][ip];
    assert( map[i] != sn );
    map[i] = sn;
    rowind[nnz] = i;
    nnz++;
  }

  new_sn_size = nnz;

  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
    for (ip=sn_size[c_sn]; ip<sn_up_size[c_sn]; ip++) {
      i = sn_rowind[c_sn][ip];
      if (map[i] != sn) { /* new row index */
	map[i] = sn;
	rowind[nnz] = i;
	nnz++;
      }
    }
  }

  for (ip=sn_size[sn]; ip<sn_up_size[sn]; ip++) {
    i = sn_rowind[sn][ip];
    if (map[i] != sn) { /* new row index */
      map[i] = sn;
      rowind[nnz] = i;
      nnz++;
    }
  }
  
  new_sn_up_size = nnz;

  if (do_order) {
    compare_indirect_map = ipostorder;
    qsort(rowind,nnz,sizeof(int),compare_indirect_ints);
  }

  /* determine whether we should merge the supernode and its children */

  {
    int n;
    double* zcount;

    n = 0;
    for (ip=0; ip<nnz; ip++) {
      i = rowind[ip];
      if (i >= n) n = i+1;
    }

    zcount = (double*) alloca(n * sizeof(double));
    assert(zcount);
    
    for (ip=0; ip<new_sn_size; ip++) {
      i = rowind[ip]; assert(i<n);
      zcount[i] = (double) (ip+1);
    }
    for (ip=new_sn_size; ip<new_sn_up_size; ip++) {
      i = rowind[ip]; assert(i<n);
      zcount[i] = (double) new_sn_size;
    }

    /*
    for (ip=0; ip<new_sn_up_size; ip++) 
      printf("row %d zcount = %.0f\n",rowind[ip],zcount[rowind[ip]]);
    */
    
    for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
      for (ip=0; ip<sn_size[c_sn]; ip++) {
	i = sn_rowind[c_sn][ip]; assert(i<n);
	zcount[i] -= (double) (ip+1);
      }
      for (ip=sn_size[c_sn]; ip<sn_up_size[c_sn]; ip++) {
	i = sn_rowind[c_sn][ip]; assert(i<n);
	zcount[i] -= (double) sn_size[c_sn];
      }
    }

    for (ip=0; ip<sn_size[sn]; ip++) {
      i = sn_rowind[sn][ip]; assert(i<n);
      zcount[i] -= (double) (ip+1);
    }
    for (ip=sn_size[sn]; ip<sn_up_size[sn]; ip++) {
      i = sn_rowind[sn][ip]; assert(i<n);
      zcount[i] -= (double) sn_size[sn];
    }

    /*
    for (ip=0; ip<new_sn_up_size; ip++) 
      printf("ROW %d zcount = %.0f\n",rowind[ip],zcount[rowind[ip]]);
    printf("zeros before merging %.0f\n",merged_znz.zeros);
    */
    
    for (ip=0; ip<new_sn_up_size; ip++) {
      i = rowind[ip]; assert(i<n);
      assert(zcount[i] >= 0.0);
      merged_znz.zeros += zcount[i];
    }

    /*printf("zeros after merging %.0f\n",merged_znz.zeros);*/

    /* voodoo constants (need some kind of a utility function */
    if ((new_sn_size < 16)
	||
	((sn_size[sn] < 50) && (merged_znz.zeros < 0.5 * merged_znz.nonzeros))
	||
	((sn_size[sn] < 250) && (merged_znz.zeros < 0.25 * merged_znz.nonzeros))
	||
	((sn_size[sn] < 500) && (merged_znz.zeros < 0.10 * merged_znz.nonzeros))
	||
	(merged_znz.zeros < 0.05 * merged_znz.nonzeros)
	) {
      /*
      taucs_printf("merging sn %d, zeros (%f) vs nonzeros (%f)\n",
		   sn,merged_znz.zeros,merged_znz.nonzeros);
      */
    } else {
      /*
      taucs_printf("sn %d, too many zeros (%f) vs nonzeros (%f)\n",
		   sn,merged_znz.zeros,merged_znz.nonzeros);
      printf("returning without merging\n");
      */
      return sn_znz;
    }
  }

  /* now merge the children lists */

  sn_size[sn]    = new_sn_size;
  sn_up_size[sn] = new_sn_up_size;
  sn_rowind[sn]  = (int*) realloc(sn_rowind[sn], 
				  new_sn_up_size * sizeof(int));
  for (ip=0; ip<new_sn_up_size; ip++) sn_rowind[sn][ip] = rowind[ip];

  /*  printf("supernode %d out of %d (merging)\n",sn,*n_sn);*/

  nchildren = 0;
  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
    for (ip=0; ip<sn_size[c_sn]; ip++) {
      i = (sn_rowind[c_sn])[ip];
      assert(column_to_sn_map[i] == c_sn);
      column_to_sn_map[i] = sn;
    }

    for (gc_sn=sn_first_child[c_sn]; gc_sn != -1; gc_sn = sn_next_child[gc_sn]) {
      rowind[nchildren] = gc_sn;
      nchildren++;
    }
  }

  /* free the children's rowind vectors */
  for (c_sn=sn_first_child[sn]; c_sn != -1; c_sn = sn_next_child[c_sn]) {
    free( sn_rowind[c_sn] );
    sn_rowind[c_sn]  = NULL;
    sn_size[c_sn]    = 0;
    sn_up_size[c_sn] = 0;
  }

  sn_first_child[sn] = -1;
  for (i=0; i<nchildren; i++) {
    sn_next_child[ rowind[i] ] = sn_first_child[sn];
    sn_first_child[sn] = rowind[i];
  }    

  /*
  printf("supernode %d out of %d (done)\n",sn,*n_sn);
  printf("returning, merging\n");
  */
  return merged_znz;
}

#ifdef TAUCS_CORE_DOUBLE

int
taucs_ccs_symbolic_elimination(taucs_ccs_matrix* A,
			       void* vL,
			       int do_order
			       )
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  int* first_child;
  int* next_child;
  int j;
  int* column_to_sn_map;
  int* map;
  int* rowind;
  int* parent;
  int* ipostorder;

  L->n           = A->n;
  /* use calloc so we can deallocate unallocated entries */
  L->sn_struct   = (int**)calloc((A->n  ),sizeof(int*)); 
  L->sn_size     = (int*) malloc((A->n+1)*sizeof(int));
  L->sn_up_size  = (int*) malloc((A->n+1)*sizeof(int));
  L->first_child = (int*) malloc((A->n+1)*sizeof(int));
  L->next_child  = (int*) malloc((A->n+1)*sizeof(int));
  
  column_to_sn_map = (int*) malloc((A->n+1)*sizeof(int));
  map              = (int*) malloc((A->n+1)*sizeof(int));
  first_child      = (int*) malloc((A->n+1)*sizeof(int));
  next_child       = (int*) malloc((A->n+1)*sizeof(int));
  parent           = (int*) malloc((A->n+1)*sizeof(int));
  rowind           = (int*) malloc((A->n  )*sizeof(int));

  if (!(L->sn_struct) || !(L->sn_size) || !(L->sn_up_size) ||
      !(L->first_child) || !(L->next_child) || !column_to_sn_map
      || !map || !first_child || !next_child || !rowind || !parent) {
    free(parent);
    free(rowind);
    free(next_child);
    free(first_child);
    free(map);
    free(column_to_sn_map);
    free(L->next_child);
    free(L->first_child);
    free(L->sn_up_size);
    free(L->sn_size);
    free(L->sn_struct);
    L->sn_struct = NULL;
    L->sn_size = L->sn_up_size = L->first_child = L->next_child = NULL;
    return -1;
  }

  if (taucs_ccs_etree(A,parent,NULL,NULL,NULL) == -1) {
    free(parent);
    free(rowind);
    free(next_child);
    free(first_child);
    free(map);
    free(column_to_sn_map);
    free(L->next_child);
    free(L->first_child);
    free(L->sn_up_size);
    free(L->sn_size);
    free(L->sn_struct);
    L->sn_struct = NULL;
    L->sn_size = L->sn_up_size = L->first_child = L->next_child = NULL;
    return -1;
  }

  if (0) {
    double wtime;
    int *cc1,*cc2,*rc1,*rc2;
    int *p1;
    int nnz1,nnz2;

    cc1=(int*)malloc((A->n)*sizeof(int));
    cc2=(int*)malloc((A->n)*sizeof(int));
    rc1=(int*)malloc((A->n)*sizeof(int));
    rc2=(int*)malloc((A->n)*sizeof(int));
    p1 =(int*)malloc((A->n)*sizeof(int));

    wtime = taucs_wtime();
    taucs_ccs_etree_liu(A,parent,cc1,rc1,&nnz1);
    wtime = taucs_wtime() - wtime;
    printf("\t\t\tLiu Analysis = %.3f seconds\n",wtime);

    wtime = taucs_wtime();
    taucs_ccs_etree(A,p1,cc2,rc2,&nnz2);
    wtime = taucs_wtime() - wtime;
    printf("\t\t\tGNP Analysis = %.3f seconds\n",wtime);

    for (j=0; j<(A->n); j++) assert(parent[j]==p1[j]);
    for (j=0; j<(A->n); j++) {
      if (cc1[j]!=cc2[j]) printf("j=%d cc1=%d cc2=%d\n",j,cc1[j],cc2[j]);
      assert(cc1[j]==cc2[j]);
    }

    for (j=0; j<(A->n); j++) {
      if (rc1[j]!=rc2[j]) printf("j=%d rc1=%d rc2=%d\n",j,rc1[j],rc2[j]);
      assert(rc1[j]==rc2[j]);
    }

    if (nnz1!=nnz2) printf("nnz1=%d nnz2=%d\n",nnz1,nnz2);
    
    free(cc1); free(cc2); free(rc1); free(rc2);
  }

  for (j=0; j <= (A->n); j++) first_child[j] = -1;
  for (j = (A->n)-1; j >= 0; j--) {
    int p = parent[j];
    next_child[j] = first_child[p];
    first_child[p] = j;
  }
  /*
  free(parent);
  ipostorder = (int*)malloc((A->n+1)*sizeof(int));
  */
  
  ipostorder = parent;
  { 
    int next = 0;
    /*int* postorder = (int*)malloc((A->n+1)*sizeof(int));*/
    recursive_postorder(A->n,first_child,next_child,
			NULL,
			ipostorder,&next);
    /*
    printf("ipostorder ");
    for (j=0; j <= (A->n); j++) printf("%d ",ipostorder[j]);
    printf("\n");
    printf(" postorder ");
    for (j=0; j <= (A->n); j++) printf("%d ",postorder[j]);
    printf("\n");
    */
  }

  L->n_sn = 0;
  for (j=0; j < (A->n); j++) map[j] = -1;
  for (j=0; j <= (A->n); j++) (L->first_child)[j] = (L->next_child)[j] = -1;
  
  if (recursive_symbolic_elimination(A->n,
				     A,
				     first_child,next_child,
				     &(L->n_sn),
				     L->sn_size,L->sn_up_size,L->sn_struct,
				     L->first_child,L->next_child,
				     rowind,
				     column_to_sn_map,
				     map,
				     do_order,ipostorder
				     )
      == -1) {
    for (j=0; j < (A->n); j++) free((L->sn_struct)[j]);

    free(parent);
    free(rowind);
    free(next_child);
    free(first_child);
    free(map);
    free(column_to_sn_map);
    free(L->next_child);
    free(L->first_child);
    free(L->sn_up_size);
    free(L->sn_size);
    free(L->sn_struct);
    L->sn_struct = NULL;
    L->sn_size = L->sn_up_size = L->first_child = L->next_child = NULL;
    return -1;
  }

  {
    double nnz   = 0.0;
    double flops = 0.0;
    int sn,i,colnnz;
    int bytes;

    bytes = 
      1*sizeof(char)                /* uplo             */
      + 2*sizeof(int)               /* n, n_sn          */
      + 3*(L->n_sn)*sizeof(int)     /* etree            */
      + 4*(L->n_sn)*sizeof(int)     /* block sizes, lda */
      + 1*(L->n_sn)*sizeof(int*)    /* row/col indices  */
      + 3*(L->n_sn)*sizeof(taucs_datatype*) /* actual blocks    */
      ;

    for (sn=0; sn<(L->n_sn); sn++) {
      bytes += (L->sn_up_size)[sn] * sizeof(int);    
      bytes += ((L->sn_size)[sn]*(L->sn_up_size)[sn]) * sizeof(taucs_datatype);

      for (i=0, colnnz = (L->sn_up_size)[sn]; 
	   i<(L->sn_size)[sn]; 
	   i++, colnnz--) {
	/* There was a bug here. I did not count muliply-adds in the
	   update part of the computation as 2 flops but one. */
	/*flops += ((double)(colnnz) - 1.0) * ((double)(colnnz) + 2.0) / 2.0;*/
	flops += 1.0 + ((double)(colnnz)) * ((double)(colnnz));
	nnz   += (double) (colnnz);
      }
    }
    taucs_printf("\t\tSymbolic Analysis of LL^T: %.2e nonzeros, %.2e flops, %.2e bytes in L\n",
		 nnz, flops, (float) bytes);
  }

  for (j=0; j < (A->n); j++) map[j] = -1;
  if (1)
  (void) recursive_amalgamate_supernodes((L->n_sn) - 1,
					 &(L->n_sn),
					 L->sn_size,L->sn_up_size,L->sn_struct,
					 L->first_child,L->next_child,
					 rowind,
					 column_to_sn_map,
					 map,
					 do_order,ipostorder
					 );


  {
    double nnz   = 0.0;
    double flops = 0.0;
    int sn,i,colnnz;
    int bytes;

    bytes = 
      1*sizeof(char)                /* uplo             */
      + 2*sizeof(int)               /* n, n_sn          */
      + 3*(L->n_sn)*sizeof(int)     /* etree            */
      + 4*(L->n_sn)*sizeof(int)     /* block sizes, lda */
      + 1*(L->n_sn)*sizeof(int*)    /* row/col indices  */
      + 3*(L->n_sn)*sizeof(taucs_datatype*) /* actual blocks    */
      ;

    for (sn=0; sn<(L->n_sn); sn++) {
      bytes += (L->sn_up_size)[sn] * sizeof(int);
      bytes += ((L->sn_size)[sn]*(L->sn_up_size)[sn]) * sizeof(taucs_datatype);

      for (i=0, colnnz = (L->sn_up_size)[sn]; 
	   i<(L->sn_size)[sn]; 
	   i++, colnnz--) {
	/* There was a bug here. I did not count muliply-adds in the
	   update part of the computation as 2 flops but one. */
	/*flops += ((double)(colnnz) - 1.0) * ((double)(colnnz) + 2.0) / 2.0;*/
	flops += 1.0 + ((double)(colnnz)) * ((double)(colnnz));
	nnz   += (double) (colnnz);
      }
    }
    taucs_printf("\t\tRelaxed  Analysis of LL^T: %.2e nonzeros, %.2e flops, %.2e bytes in L\n",
		 nnz, flops, (float) bytes);
  }

  /*
  {
    int i;
    printf("c2sn: ");
    for (i=0; i<A->n; i++) printf("%d ",column_to_sn_map[i]);
    printf("\n");
  }
  */
  
  free(parent);
  free(rowind);
  free(map);
  free(column_to_sn_map);
  free(next_child);
  free(first_child);

  L->sn_blocks_ld  = malloc((L->n_sn) * sizeof(int));
  L->sn_blocks     = calloc((L->n_sn), sizeof(taucs_datatype*)); /* so we can free before allocation */
  
  L->up_blocks_ld  = malloc((L->n_sn) * sizeof(int));
  L->up_blocks     = calloc((L->n_sn), sizeof(taucs_datatype*));

  if (!(L->sn_blocks_ld)
      || !(L->sn_blocks_ld)
      || !(L->sn_blocks)
      || !(L->up_blocks_ld)
      || !(L->up_blocks))
    return -1; /* the caller will free L */

  return 0;
}
#endif

/*************************************************************/
/* factor routines                                           */
/*************************************************************/

static supernodal_frontal_matrix*
recursive_multifrontal_supernodal_factor_llt(int sn,       /* this supernode */
					     int is_root,  /* is v the root? */
					     int* bitmap,
					     taucs_ccs_matrix* A,
					     supernodal_factor_matrix* snL,
					     int* fail)
{
  supernodal_frontal_matrix* my_matrix=NULL;
  supernodal_frontal_matrix* child_matrix=NULL;
  int child;
  int* v; 
  int  sn_size;
  int* first_child   = snL->first_child;
  int* next_child    = snL->next_child;

  v = &( snL->sn_struct[sn][0] );

  if (!is_root)
    sn_size = snL->sn_size[sn];
  else
    sn_size = -1;

  for (child = first_child[sn]; child != -1; child = next_child[child]) {
    child_matrix = 
      recursive_multifrontal_supernodal_factor_llt(child,
						   FALSE,
						   bitmap,
						   A,snL,fail);
    if (*fail) { 
      if (my_matrix) supernodal_frontal_free(my_matrix);
      return NULL;
    }

    if (!is_root) {
      if (!my_matrix) {
	my_matrix =  supernodal_frontal_create(v,sn_size,
					       snL->sn_up_size[sn],
					       snL->sn_struct[sn]);
	if (!my_matrix) {
	  *fail = TRUE;
	  supernodal_frontal_free(child_matrix);
	  return NULL;
	}
      }
      multifrontal_supernodal_front_extend_add(my_matrix,child_matrix,bitmap);
    }
    /* moved outside "if !is_root"; Sivan 27 Feb 2002 */
    supernodal_frontal_free(child_matrix);
  }

  /* in case we have no children, we allocate now */
  if (!is_root && !my_matrix) {
    my_matrix =  supernodal_frontal_create(v,sn_size,
					   snL->sn_up_size[sn],
					   snL->sn_struct[sn]);
    if (!my_matrix) {
      *fail = TRUE;
      return NULL;
    }
  }
  
  if(!is_root) {
    if (multifrontal_supernodal_front_factor(sn,
					     v,sn_size,
					     A,
					     my_matrix,
					     bitmap,
					     snL)) { 
      /* nonpositive pivot */
      *fail = TRUE;
      supernodal_frontal_free(my_matrix);
      return NULL;
    }
  }
  return my_matrix;
}

void* 
taucs_dtl(ccs_factor_llt_mf)(taucs_ccs_matrix* A)
{
  supernodal_factor_matrix* L;
  int* map;
  int fail;
  double wtime, ctime;

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  L = multifrontal_supernodal_create();
  if (!L) return NULL;

#ifdef TAUCS_CORE_COMPLEX
  fail = taucs_ccs_symbolic_elimination(A,L,
					TRUE /* sort, to avoid complex conjuation */ );
#else
  fail = taucs_ccs_symbolic_elimination(A,L,
					FALSE /* don't sort row indices */ );
#endif
  if (fail == -1) {
    taucs_supernodal_factor_free(L);
    return NULL;
  }

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSymbolic Analysis            = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);

  map = (int*)malloc((A->n+1)*sizeof(int));
  if (!map) {
    taucs_supernodal_factor_free(L);
    return NULL;
  }

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  fail = FALSE;
  recursive_multifrontal_supernodal_factor_llt((L->n_sn),  
					       TRUE, 
					       map,
					       A,L,&fail);

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSupernodal Multifrontal LL^T = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);

  free(map);

  if (fail) {
    taucs_supernodal_factor_free(L);
    return NULL;
  }

  return (void*) L;
}

/*************************************************************/
/* symbolic-numeric routines                                 */
/*************************************************************/

void* 
taucs_dtl(ccs_factor_llt_symbolic)(taucs_ccs_matrix* A)
{
  supernodal_factor_matrix* L;
  int fail;
  double wtime, ctime;

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  L = multifrontal_supernodal_create();
  if (!L) return NULL;

#ifdef TAUCS_CORE_COMPLEX
  fail = taucs_ccs_symbolic_elimination(A,L,
					TRUE /* sort, to avoid complex conjuation */ );
#else
  fail = taucs_ccs_symbolic_elimination(A,L,
					FALSE /* don't sort row indices */ );
#endif

  if (fail == -1) {
    taucs_supernodal_factor_free(L);
    return NULL;
  }

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSymbolic Analysis            = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);
  return L;
}

int 
taucs_dtl(ccs_factor_llt_numeric)(taucs_ccs_matrix* A,void* vL)
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  int* map;
  int fail;
  double wtime, ctime;

  map = (int*)malloc((A->n+1)*sizeof(int));

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  fail = FALSE;
  recursive_multifrontal_supernodal_factor_llt((L->n_sn),  
					       TRUE, 
					       map,
					       A,L,&fail);

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSupernodal Multifrontal LL^T = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);

  free(map);

  if (fail) {
    taucs_supernodal_factor_free_numeric(L);
    return -1;
  }

  return 0;
}

/*************************************************************/
/* left-looking factor routines                              */
/*************************************************************/

static void
recursive_leftlooking_supernodal_update(int J,int K,
					int bitmap[],
					taucs_datatype* dense_update_matrix,
					taucs_ccs_matrix* A,
					supernodal_factor_matrix* L)
{
  int i,j,ir;
  int  child;
  int* first_child   = L->first_child;
  int* next_child    = L->next_child;
  int sn_size_father = (L->sn_size)[J];
  int sn_up_size_father = (L->sn_up_size)[J];
  int sn_size_child = (L->sn_size)[K];
  int sn_up_size_child = (L->sn_up_size)[K];
  int exist_upd=0;
  int first_row;
  int row_count=0;
  int PK,M,N,LDA,LDB,LDC;

  for(i=0;i<sn_size_father;i++) {
    bitmap[L->sn_struct[J][i]]=i+1;
  }

  for(i=sn_size_father;i<sn_up_size_father;i++)
    bitmap[L->sn_struct[J][i]] = i - sn_size_father + 1;

  for(i=sn_size_child;i<sn_up_size_child;i++)
    /* is this row index included in the columns of sn J? */
    if(bitmap[L->sn_struct[K][i]]
       && L->sn_struct[K][i] <= L->sn_struct[J][sn_size_father-1]) {
      if(!exist_upd) first_row = i;
      row_count++;
      exist_upd = 1;
      /*taucs_printf("update from K = %d to J = %d \n",K,J);*/
      /* loop over columns of sn K */
            
      /* for(j=0;j<sn_size_child;j++)
	for(ir=i;ir<sn_up_size_child;ir++)
	  if( L->sn_struct[K][ir] <= L->sn_struct[J][sn_size_father-1]){
	    L->sn_blocks[J][ (bitmap[L->sn_struct[K][i]]-1)*(L->sn_blocks_ld[J])+(bitmap[L->sn_struct[K][ir]]-1)] -= L->up_blocks[K][j*(L->up_blocks_ld[K])+ir-sn_size_child]* L->up_blocks[K][j*L->up_blocks_ld[K]+i-sn_size_child];
	    taucs_printf("sn_block: L[%d,%d] = %lf\n",(bitmap[L->sn_struct[K][ir]]-1),(bitmap[L->sn_struct[K][i]]-1),L->sn_blocks[J][ (bitmap[L->sn_struct[K][i]]-1)*(L->sn_blocks_ld[J])+(bitmap[L->sn_struct[K][ir]]-1)]);}
	  else{
	    L->up_blocks[J][ (bitmap[L->sn_struct[K][i]]-1)*(L->up_blocks_ld[J])+(bitmap[L->sn_struct[K][ir]]-1)] -=  L->up_blocks[K][j*L->up_blocks_ld[K]+ir-sn_size_child]* L->up_blocks[K][j*L->up_blocks_ld[K]+i-sn_size_child];
	   taucs_printf("up_block: L[%d,%d] = %lf\n",(bitmap[L->sn_struct[K][ir]]-1),(bitmap[L->sn_struct[K][i]]-1),L->up_blocks[J][ (bitmap[L->sn_struct[K][i]]-1)*(L->up_blocks_ld[J])+(bitmap[L->sn_struct[K][ir]]-1)]);
	   }*/
        }

  if(exist_upd){
    LDA = LDB = (L->up_blocks_ld)[K];
    M  = sn_up_size_child - first_row ; /* +-1 ? */    
    LDC =  sn_up_size_father;
    N  = row_count; 
    PK = L->sn_size[K];    

    taucs_gemm ("No Conjugate",
		"Conjugate",
		&M,&N,&PK,
		&taucs_one_const,
		&(L->up_blocks[K][first_row-sn_size_child]),&LDA,
		&(L->up_blocks[K][first_row-sn_size_child]),&LDB,
		&taucs_zero_const,
		dense_update_matrix,&LDC);

    /*for(j=0;j<row_count;j++)
       for(ir=0;ir<sn_up_size_father;ir++)
	 taucs_printf("dense[%d,%d] = %lf\n",ir,j,dense_update_matrix[j*LDC+ir]);
    */

    for(j=0;j<row_count;j++)
      for(ir=j;ir<row_count;ir++){

#if 0
	L->sn_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*sn_size_father+(bitmap[L->sn_struct[K][first_row+ir]]-1)] -= dense_update_matrix[j*LDC+ir];
#endif

	L->sn_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*sn_size_father+(bitmap[L->sn_struct[K][first_row+ir]]-1)] = 
	  taucs_sub( L->sn_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*sn_size_father+(bitmap[L->sn_struct[K][first_row+ir]]-1)] , dense_update_matrix[j*LDC+ir]);

	/*	taucs_printf("sn_block: L[%d,%d] = %lf\n",(bitmap[L->sn_struct[K][first_row+ir]]-1),(bitmap[L->sn_struct[K][first_row+j]]-1),L->sn_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*sn_size_father+(bitmap[L->sn_struct[K][first_row+ir]]-1)]);*/

      }

    for(j=0;j<row_count;j++)
      for(ir=row_count;ir<M;ir++){
#if 0
	L->up_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*(L->up_blocks_ld)[J]+(bitmap[L->sn_struct[K][ir+first_row]]-1)] -= dense_update_matrix[j*LDC+ir];
#endif

	L->up_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*(L->up_blocks_ld)[J]+(bitmap[L->sn_struct[K][ir+first_row]]-1)] =
	  taucs_sub( L->up_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*(L->up_blocks_ld)[J]+(bitmap[L->sn_struct[K][ir+first_row]]-1)] , dense_update_matrix[j*LDC+ir]);

	/*	taucs_printf("up_block: L[%d,%d] = %lf\n",(bitmap[L->sn_struct[K][ir+first_row]]-1),(bitmap[L->sn_struct[K][first_row+j]]-1),L->up_blocks[J][(bitmap[L->sn_struct[K][first_row+j]]-1)*(L->up_blocks_ld)[J]+(bitmap[L->sn_struct[K][ir+first_row]]-1)]);*/

	}
    for(i=0;i<sn_up_size_father;i++)
      bitmap[L->sn_struct[J][i]]=0;
    
    for (child = first_child[K]; child != -1; child = next_child[child]) {
      recursive_leftlooking_supernodal_update(J,child,
					      bitmap,dense_update_matrix,
					      A,L);
    }
  }
  else
    for(i=0;i<sn_up_size_father;i++)
      bitmap[L->sn_struct[J][i]]=0;

}

static int
leftlooking_supernodal_front_factor(int sn,
				    int* bitmap,
				    taucs_ccs_matrix* A,
				    supernodal_factor_matrix* L)
{
  int ip,jp;
  int*    ind;
  taucs_datatype* re;
  int INFO;

  int sn_size = (L->sn_size)[sn];
  int up_size = (L->sn_up_size)[sn] - (L->sn_size)[sn];

  /* creating transform for real indices */
  for(ip=0;ip<(L->sn_up_size)[sn];ip++) bitmap[(L->sn_struct)[sn][ip]] = ip;

  for(jp=0;jp<sn_size;jp++) {
    ind = &(A->rowind[A->colptr[ (L->sn_struct)[sn][jp] ]]);
    re  = &(A->taucs_values[A->colptr[ (L->sn_struct)[sn][jp] ]]); 
    for(ip=0;
	ip < A->colptr[ (L->sn_struct)[sn][jp] + 1 ] 
           - A->colptr[ (L->sn_struct)[sn][jp] ];
	ip++) {
      if (bitmap[ind[ip]] < sn_size)
	(L->sn_blocks)[sn][ (L->sn_blocks_ld)[sn]*jp + bitmap[ind[ip]]] =
	  taucs_add( (L->sn_blocks)[sn][ (L->sn_blocks_ld)[sn]*jp + bitmap[ind[ip]]] , re[ip] );
      else
	(L->up_blocks)[sn][ (L->up_blocks_ld)[sn]*jp + bitmap[ind[ip]] - sn_size] =
	taucs_add( (L->up_blocks)[sn][ (L->up_blocks_ld)[sn]*jp + bitmap[ind[ip]] - sn_size] , re[ip] );
    }
  }
  
  /* we use the BLAS through the Fortran interface */

  /* solving of lower triangular system for L */
  if (sn_size)
    taucs_potrf ("LOWER",
		 &sn_size,
		 (L->sn_blocks)[sn],&((L->sn_blocks_ld)[sn]),
		 &INFO);

  if (INFO) {
    taucs_printf("\t\tLL^T Factorization: Matrix is not positive definite.\n");
    taucs_printf("\t\t                    nonpositive pivot in column %d\n",
		 (L->sn_struct)[INFO-1]);
    return -1;
  }

  /* getting completion for found columns of L */
  if (up_size && sn_size)
    taucs_trsm ("Right",
		"Lower",
		"Conjugate",
		"No unit diagonal",
		&up_size,&sn_size,
		&taucs_one_const,
		(L->sn_blocks)[sn],&((L->sn_blocks_ld)[sn]),
		(L->up_blocks)[sn],&((L->up_blocks_ld)[sn]));

  return 0;
}

static int
recursive_leftlooking_supernodal_factor_llt(int sn,       /* this supernode */
					    int is_root,  /* is v the root? */
					    int* bitmap,
					    int* indmap,
					    taucs_ccs_matrix* A,
					    supernodal_factor_matrix* L)
{
  int  child;
  int  sn_size;
  int* first_child   = L->first_child;
  int* next_child    = L->next_child;
  taucs_datatype* dense_update_matrix = NULL;

  if (!is_root)
    sn_size = L->sn_size[sn];
  else
    sn_size = -1;

  if (!is_root) { 
    (L->sn_blocks   )[sn] = (taucs_datatype*)calloc(((L->sn_size)[sn])
					    *((L->sn_size)[sn]),sizeof(taucs_datatype));
    (L->sn_blocks_ld)[sn] = (L->sn_size   )[sn];

    (L->up_blocks   )[sn] = (taucs_datatype*)calloc(((L->sn_up_size)[sn]-(L->sn_size)[sn])
					    *((L->sn_size)[sn]),sizeof(taucs_datatype));
    (L->up_blocks_ld)[sn] = (L->sn_up_size)[sn]-(L->sn_size)[sn];

    if (!((L->sn_blocks)[sn]) || !((L->up_blocks)[sn])) 
      return -1; /* the caller will free L */
  }

  for (child = first_child[sn]; child != -1; child = next_child[child]) {
    if (recursive_leftlooking_supernodal_factor_llt(child,
						    FALSE,
						    bitmap,
						    indmap,
						    A,L)
	== -1 ) {
      free(dense_update_matrix);
      return -1;
    }
    
    if (!is_root) {
      if (!dense_update_matrix) {
	dense_update_matrix = 
	  (taucs_datatype*) calloc((L->sn_up_size)[sn]*(L->sn_size)[sn],sizeof(taucs_datatype));
	if (!dense_update_matrix) return -1; /* caller will free L */
      }
      recursive_leftlooking_supernodal_update(sn,child,
					      bitmap,dense_update_matrix,
					      A,L);
    }
  }
  free(dense_update_matrix);
  
  if(!is_root) {
    if (leftlooking_supernodal_front_factor(sn,
					    indmap,
					    A,
					    L)) {
      return -1; /* nonpositive pivot */
    }
  }

  return 0;
}

void* 
taucs_dtl(ccs_factor_llt_ll)(taucs_ccs_matrix* A)
{
  supernodal_factor_matrix* L;
  int* map;
  int *map2;
  double wtime, ctime;
  int fail;

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  L = multifrontal_supernodal_create();
  if (!L) return NULL;

  fail = taucs_ccs_symbolic_elimination(A,L,
					TRUE /* sort row indices */);

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSymbolic Analysis            = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);

  map  = (int*)malloc((A->n+1)*sizeof(int));
  map2 = (int*)calloc((A->n+1),sizeof(int));

  if (fail == -1 || !map || !map2) {
    taucs_supernodal_factor_free(L);
    free(map2);
    free(map);
    return NULL;
  }

  wtime = taucs_wtime();
  ctime = taucs_ctime();

  if (recursive_leftlooking_supernodal_factor_llt((L->n_sn),  
						  TRUE, 
						  map2,
						  map,
						  A,L)
      == -1) {
    taucs_supernodal_factor_free(L);
    free(map);
    free(map2);
    return NULL;
  }

  wtime = taucs_wtime()-wtime;
  ctime = taucs_ctime()-ctime;
  taucs_printf("\t\tSupernodal Left-Looking LL^T = % 10.3f seconds (%.3f cpu)\n",
	       wtime,ctime);

  free(map);
  free(map2);

  return (void*) L;
}

/*************************************************************/
/* supernodal solve routines                                 */
/*************************************************************/

static void 
recursive_supernodal_solve_l(int sn,       /* this supernode */
			     int is_root,  /* is v the root? */
			     int* first_child, int* next_child,
			     int** sn_struct, int* sn_sizes, int* sn_up_sizes,
			     int* sn_blocks_ld,taucs_datatype* sn_blocks[],
			     int* up_blocks_ld,taucs_datatype* up_blocks[],
			     taucs_datatype x[], taucs_datatype b[],
			     taucs_datatype t[])
{
  int child;
  int  sn_size; /* number of rows/columns in the supernode    */
  int  up_size; /* number of rows that this supernode updates */
  int    ione = 1;

  taucs_datatype* xdense;
  taucs_datatype* bdense;
  double  flops;
  int i,j,ip,jp;

  for (child = first_child[sn]; child != -1; child = next_child[child]) {
    recursive_supernodal_solve_l(child,
				 FALSE,
				 first_child,next_child,
 				 sn_struct,sn_sizes,sn_up_sizes,
				 sn_blocks_ld,sn_blocks,
				 up_blocks_ld,up_blocks,
				 x,b,t);
  }

  if(!is_root) {

    sn_size = sn_sizes[sn];
    up_size = sn_up_sizes[sn] - sn_sizes[sn];

    flops = ((double)sn_size)*((double)sn_size) 
      + 2.0*((double)sn_size)*((double)up_size);

    if (flops > BLAS_FLOPS_CUTOFF) {
      xdense = t;
      bdense = t + sn_size;
      
      for (i=0; i<sn_size; i++)
	xdense[i] = b[ sn_struct[ sn ][ i ] ];
      for (i=0; i<up_size; i++)
	bdense[i] = taucs_zero;
      
      taucs_trsm ("Left",
		  "Lower",
		  "No Conjugate",
		  "No unit diagonal",
		  &sn_size,&ione,
		  &taucs_one_const,
		  sn_blocks[sn],&(sn_blocks_ld[sn]),
		  xdense       ,&sn_size);
      
      if (up_size > 0 && sn_size > 0) {
	taucs_gemm ("No Conjugate","No Conjugate",
		    &up_size, &ione, &sn_size,
		    &taucs_one_const,
		    up_blocks[sn],&(up_blocks_ld[sn]),
		    xdense       ,&sn_size,
		    &taucs_zero_const,
		    bdense       ,&up_size);
      }
      
      for (i=0; i<sn_size; i++)
	x[ sn_struct[ sn][ i ] ]  = xdense[i];
      for (i=0; i<up_size; i++)
	/*b[ sn_struct[ sn ][ sn_size + i ] ] -= bdense[i];*/
	b[ sn_struct[ sn ][ sn_size + i ] ] =
	  taucs_sub( b[ sn_struct[ sn ][ sn_size + i ] ] , bdense[i] );

#if 1
    }
#else
    } else if (sn_size > SOLVE_DENSE_CUTOFF) {

      xdense = t;
      bdense = t + sn_size;
      
      for (i=0; i<sn_size; i++)
	xdense[i] = b[ sn_struct[ sn ][ i ] ];
      for (i=0; i<up_size; i++)
	bdense[i] = 0.0;
      
      for (jp=0; jp<sn_size; jp++) {
	xdense[jp] = xdense[jp] / sn_blocks[sn][ sn_blocks_ld[sn]*jp + jp];

	for (ip=jp+1; ip<sn_size; ip++) {
	  xdense[ip] -= xdense[jp] * sn_blocks[sn][ sn_blocks_ld[sn]*jp + ip];
	}
      }

      for (jp=0; jp<sn_size; jp++) {
	for (ip=0; ip<up_size; ip++) {
	  bdense[ip] += xdense[jp] * up_blocks[sn][ up_blocks_ld[sn]*jp + ip];
	}
      }

      for (i=0; i<sn_size; i++)
	x[ sn_struct[ sn][ i ] ]  = xdense[i];
      for (i=0; i<up_size; i++)
	b[ sn_struct[ sn ][ sn_size + i ] ] -= bdense[i];
      
    } else {

      for (jp=0; jp<sn_size; jp++) {
	j = sn_struct[sn][jp];
	x[j] = b[j] / sn_blocks[sn][ sn_blocks_ld[sn]*jp + jp];
	for (ip=jp+1; ip<sn_size; ip++) {
	  i = sn_struct[sn][ip];
	  b[i] -= x[j] * sn_blocks[sn][ sn_blocks_ld[sn]*jp + ip];
	}

	for (ip=0; ip<up_size; ip++) {
	  i = sn_struct[sn][sn_size + ip];
	  b[i] -= x[j] * up_blocks[sn][ up_blocks_ld[sn]*jp + ip];
	}
      }

    }
#endif
  }
}

static void 
recursive_supernodal_solve_lt(int sn,       /* this supernode */
 			      int is_root,  /* is v the root? */
			      int* first_child, int* next_child,
			      int** sn_struct, int* sn_sizes, int* sn_up_sizes,
			      int* sn_blocks_ld,taucs_datatype* sn_blocks[],
			      int* up_blocks_ld,taucs_datatype* up_blocks[],
			      taucs_datatype x[], taucs_datatype b[],
			      taucs_datatype t[])
{
  int child;
  int  sn_size; /* number of rows/columns in the supernode    */
  int  up_size; /* number of rows that this supernode updates */
  int    ione = 1;

  taucs_datatype* xdense;
  taucs_datatype* bdense;
  double  flops;
  int i,j,ip,jp;

  if(!is_root) {

    sn_size = sn_sizes[sn];
    up_size = sn_up_sizes[sn]-sn_sizes[sn];
    
    flops = ((double)sn_size)*((double)sn_size) 
      + 2.0*((double)sn_size)*((double)up_size);

    if (flops > BLAS_FLOPS_CUTOFF) {

      bdense = t;
      xdense = t + sn_size;
      
      for (i=0; i<sn_size; i++)
	bdense[i] = b[ sn_struct[ sn][ i ] ];
      for (i=0; i<up_size; i++)
	xdense[i] = x[ sn_struct[sn][sn_size+i] ];
      
      if (up_size > 0 && sn_size > 0)
	taucs_gemm ("Conjugate","No Conjugate",
		     &sn_size, &ione, &up_size,
		     &taucs_minusone_const,
		     up_blocks[sn],&(up_blocks_ld[sn]),
		     xdense       ,&up_size,
		     &taucs_one_const,
		     bdense       ,&sn_size);
      
      taucs_trsm ("Left",
		  "Lower",
		  "Conjugate",
		  "No unit diagonal",
		  &sn_size,&ione,
		  &taucs_one_const,
		  sn_blocks[sn],&(sn_blocks_ld[sn]),
		  bdense       ,&sn_size);
      
      for (i=0; i<sn_size; i++)
	x[ sn_struct[ sn][ i ] ]  = bdense[i];
#if 1
    }
#else    
    } else if (sn_size > SOLVE_DENSE_CUTOFF) {

      bdense = t;
      xdense = t + sn_size;
      
      for (i=0; i<sn_size; i++)
	bdense[i] = b[ sn_struct[ sn][ i ] ];
      for (i=0; i<up_size; i++)
	xdense[i] = x[ sn_struct[sn][sn_size+i] ];
      
      for (ip=sn_size-1; ip>=0; ip--) {
	for (jp=0; jp<up_size; jp++) {
	  bdense[ip] -= xdense[jp] * up_blocks[sn][ up_blocks_ld[sn]*ip + jp];
	}
      }

      for (ip=sn_size-1; ip>=0; ip--) {
	for (jp=sn_size-1; jp>ip; jp--) {
	  bdense[ip] -= bdense[jp] * sn_blocks[sn][ sn_blocks_ld[sn]*ip + jp];
	}
	bdense[ip] = bdense[ip] / sn_blocks[sn][ sn_blocks_ld[sn]*ip + ip];
      }

      for (i=0; i<sn_size; i++)
	x[ sn_struct[ sn][ i ] ]  = bdense[i];
    
    } else {

      for (ip=sn_size-1; ip>=0; ip--) {
	i = sn_struct[sn][ip];

	for (jp=0; jp<up_size; jp++) {
	  j = sn_struct[sn][sn_size + jp];
	  b[i] -= x[j] * up_blocks[sn][ up_blocks_ld[sn]*ip + jp];
	}

	for (jp=sn_size-1; jp>ip; jp--) {
	  j = sn_struct[sn][jp];
	  b[i] -= x[j] * sn_blocks[sn][ sn_blocks_ld[sn]*ip + jp];
	}
	x[i] = b[i] / sn_blocks[sn][ sn_blocks_ld[sn]*ip + ip];

      }

    }
#endif

  }

  for (child = first_child[sn]; child != -1; child = next_child[child]) {
    recursive_supernodal_solve_lt(child,
				  FALSE,
				  first_child,next_child,
				  sn_struct,sn_sizes,sn_up_sizes,
				  sn_blocks_ld,sn_blocks,
				  up_blocks_ld,up_blocks,
				  x,b,t);
  }
}

int 
taucs_dtl(supernodal_solve_llt)(void* vL, void* vx, void* vb)
{
  supernodal_factor_matrix* L = (supernodal_factor_matrix*) vL;
  taucs_datatype* x = (taucs_datatype*) vx;
  taucs_datatype* b = (taucs_datatype*) vb;
  taucs_datatype* y;
  taucs_datatype* t; /* temporary vector */
  int     i;
  
  y = malloc((L->n) * sizeof(taucs_datatype));
  t = malloc((L->n) * sizeof(taucs_datatype));
  if (!y || !t) {
    free(y);
    free(t);
    taucs_printf("multifrontal_supernodal_solve_llt: out of memory\n");
    return -1;
  }

  for (i=0; i<L->n; i++) x[i] = b[i];

  recursive_supernodal_solve_l (L->n_sn,
				TRUE,  /* this is the root */
				L->first_child, L->next_child,
				L->sn_struct,L->sn_size,L->sn_up_size,
				L->sn_blocks_ld, L->sn_blocks,
				L->up_blocks_ld, L->up_blocks,
				y, x, t);

  recursive_supernodal_solve_lt(L->n_sn,
				TRUE,  /* this is the root */
				L->first_child, L->next_child,
				L->sn_struct,L->sn_size,L->sn_up_size,
				L->sn_blocks_ld, L->sn_blocks,
				L->up_blocks_ld, L->up_blocks,
				x, y, t);

  free(y);
  free(t);
    
  return 0;
}

/*************************************************************/
/* generic interfaces to user-callable routines              */
/*************************************************************/

#ifdef TAUCS_CORE_DOUBLE

void* taucs_ccs_factor_llt_mf(taucs_ccs_matrix* A)
{
  if (A->flags & TAUCS_DOUBLE)
    return taucs_dccs_factor_llt_mf(A);

  if (A->flags & TAUCS_SINGLE)
    return taucs_sccs_factor_llt_mf(A);

  if (A->flags & TAUCS_DCOMPLEX)
    return taucs_zccs_factor_llt_mf(A);

  if (A->flags & TAUCS_SCOMPLEX)
    return taucs_cccs_factor_llt_mf(A);

  assert(0);
  return NULL;
}

void* taucs_ccs_factor_llt_ll(taucs_ccs_matrix* A)
{
  if (A->flags & TAUCS_DOUBLE)
    return taucs_dccs_factor_llt_ll(A);

  if (A->flags & TAUCS_SINGLE)
    return taucs_sccs_factor_llt_ll(A);

  if (A->flags & TAUCS_DCOMPLEX)
    return taucs_zccs_factor_llt_ll(A);

  if (A->flags & TAUCS_SCOMPLEX)
    return taucs_cccs_factor_llt_ll(A);

  assert(0);
  return NULL;
}

void* taucs_ccs_factor_llt_symbolic(taucs_ccs_matrix* A)
{
  if (A->flags & TAUCS_DOUBLE)
    return taucs_dccs_factor_llt_symbolic(A);

  if (A->flags & TAUCS_SINGLE)
    return taucs_sccs_factor_llt_symbolic(A);

  if (A->flags & TAUCS_DCOMPLEX)
    return taucs_zccs_factor_llt_symbolic(A);

  if (A->flags & TAUCS_SCOMPLEX)
    return taucs_cccs_factor_llt_symbolic(A);

  assert(0);
  return NULL;
}

int taucs_ccs_factor_llt_numeric(taucs_ccs_matrix* A, void* L)
{
  if (A->flags & TAUCS_DOUBLE)
    return taucs_dccs_factor_llt_numeric(A,L);

  if (A->flags & TAUCS_SINGLE)
    return taucs_sccs_factor_llt_numeric(A,L);

  if (A->flags & TAUCS_DCOMPLEX)
    return taucs_zccs_factor_llt_numeric(A,L);

  if (A->flags & TAUCS_SCOMPLEX)
    return taucs_cccs_factor_llt_numeric(A,L);

  assert(0);
  return -1;
}

int taucs_supernodal_solve_llt(void* L, void* x, void* b)
{
  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DOUBLE)
    return taucs_dsupernodal_solve_llt(L,x,b);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SINGLE)
    return taucs_ssupernodal_solve_llt(L,x,b);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DCOMPLEX)
    return taucs_zsupernodal_solve_llt(L,x,b);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SCOMPLEX)
    return taucs_csupernodal_solve_llt(L,x,b);

  assert(0);
  return -1;
}

void taucs_supernodal_factor_free(void* L)
{
  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DOUBLE) {
    taucs_dsupernodal_factor_free(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SINGLE) {
    taucs_ssupernodal_factor_free(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DCOMPLEX) {
    taucs_zsupernodal_factor_free(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SCOMPLEX) {
    taucs_csupernodal_factor_free(L);
    return;
  }

  assert(0);
}

void taucs_supernodal_factor_free_numeric(void* L)
{
  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DOUBLE) {
    taucs_dsupernodal_factor_free_numeric(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SINGLE) {
    taucs_ssupernodal_factor_free_numeric(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DCOMPLEX) {
    taucs_zsupernodal_factor_free_numeric(L);
    return;
  }

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SCOMPLEX) {
    taucs_csupernodal_factor_free_numeric(L);
    return;
  }

  assert(0);
}

taucs_ccs_matrix* 
taucs_supernodal_factor_to_ccs(void* L)
{
  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DOUBLE)
    return taucs_dsupernodal_factor_to_ccs(L);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SINGLE)
    return taucs_ssupernodal_factor_to_ccs(L);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_DCOMPLEX)
    return taucs_zsupernodal_factor_to_ccs(L);

  if (((supernodal_factor_matrix*) L)->flags & TAUCS_SCOMPLEX)
    return taucs_csupernodal_factor_to_ccs(L);

  assert(0);
  return NULL;
}

#endif

/*************************************************************/
/* end of file                                               */
/*************************************************************/













